/*
** Module   : ScrollableSelectionListGuiImpl.java
** Abstract : scrollable selection list implementation 
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150423 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 EVL 20150514 Changing scrollbar painting approach.  Several minor UI fixes.  Adding method
**                  to help dismissing of the drop-down on outside mouse press.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.Position;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Scrollable selection list GUI implementation.  To be used to support combo box and selection
 * list widgets. 
 */
@SuppressWarnings("unchecked")
public class ScrollableSelectionListGuiImpl
extends ScrollableList<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Widget configuration */
   private BaseConfig config;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Row height */
   private int rowHeight;

   /** The vertical scroll bar associated with this scrollable list. */
   private ScrollBar<GuiOutputManager> vbar;

   /** Selected item box color. */
   private ColorRgb selectionBack;

   /** Selected item text color. */
   private ColorRgb selectionFore;

   /** Disabled state text color. */
   private ColorRgb disabledFore;
   
   /** The curent top row in the list to draw. */
   private int rowTop;

   /**
    * Create component using given configuration and data model.
    * 
    * @param    box
    *           The instance with which this object must be associated.
    * @param    model
    *           The data model that represents the contents of the list.
    */
   ScrollableSelectionListGuiImpl(ComboBox<GuiOutputManager> box, ListModel<String> model)
   {
      super(box, model);
      // initialize screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // and font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, box.config(), false);
      // color initializing
      selectionBack = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      selectionFore = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      disabledFore  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      // top row number will be initialized later in setTopRow()
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      super.requestFocus();
   }

   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public void draw()
   {
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = box.config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_WINDOW, SysColor.COLOR_BTNTEXT);

      // determine colors to use
      gc.colors();

      // get actual rows to view
      final int rowsVisible = visibleRows();
      // set the position and range for the scrollbar
      if (rowsVisible < model().size())
      {
         vbar._setVisible(true);
         vbar.setMax(model().size());
         vbar.setVisibleRange(rowsVisible);
         vbar.setPosition(rowTop);
      }
      else
      {
         vbar._setVisible(false);
      }
      
      // resolve font
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final int rowHeight = screen().coordinates().heightToNative(1.0); 
      d.height = rowsVisible * rowHeight;

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);

      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);
            
            // fill the background
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);

            // draw entry field text
            gd.setGuiFont(fntCache.font);
            gd.setFontStyle(fntCache.font.style);
            
            int x = po.x;
            int y = po.y;
            int fontSize = fntCache.font.pointSize;
            // adjust top row number to have last item from the list as bottom number 
            if ((rowTop + rowsVisible) > model().size())
            {
               rowTop = model().size() - rowsVisible; 
            }
            
            // check if the list is enabled
            boolean enabled = isEnabled();
            
            // draw items from the top to bottom
            for (int i = rowTop; i < model().size() && i < rowsVisible + rowTop;
                            i++, y+= rowHeight)
            {
               if (y < 0)
                  continue;
               
               // draw nex visible item in the list          
               drawItem(i, x, y, d.width, rowHeight, fontSize, enabled);
            }
         }
         
      });
   }

   /**
    * Process scroll event.
    * 
    * @param   event
    *          Scroll event.
    */
   @Override
   public void processScrollEvent(ScrollEvent event)
   {
      // find out if the key or mouse scrolling
      boolean mouseOrig = event.isMouseOrigin();
      
      // get the new row to be set as current in the list
      // mouse origin scrollng has special handling
      if (mouseOrig)
      {
         // determine scrolling direction
         ScrollDirection dir = event.direction();
         // change top row respectively
         if (ScrollDirection.up(dir) && rowTop > 0)
         {
            rowTop -= 1;
         }
         else if (ScrollDirection.down(dir) && (rowTop + visibleRows()) < model().size())
         {
            rowTop += 1;
         }
      }
      else // key scrolling
      {
         super.ensureRowVisible((int) event.limit().y, false);
      
         // recalc top row changes
         if (currentRow() < rowTop)
         {
            // upper than visible region
            rowTop = currentRow(); 
         }
         else if (currentRow() >= rowTop + visibleRows())
         {
            // lower than visible region
            rowTop = currentRow() - visibleRows() + 1;
         }

         // and notify main combo-box about new selection for keyboard event origin case
         selectItem();
      }
      
      // pass the rest to scroll listeners for additional pcocessing
      super.processScrollEvent(event);
      
      // and repaint the list
      ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
      {
         @Override
         public void run()
         {
            repaint();
         }
      });
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      int rowHeight = screen().coordinates().heightToNative(1.0);
      
      synchronized (this)
      {
         if (model().size() > 0)
         {
            NativePoint n = translate(this, e);
            
            // calculate clicked line
            int row = n.y / rowHeight + rowTop;

            // in combo-box it is possible to generate selection for already selected item
            // go to selected item
            super.ensureRowVisible(row, false);
            // select new item
            selectItem();
            // and generate LEAVE to dismiss drop-down in respective mode
            if (getBox().config().mode != ComboBoxConfig.Mode.SIMPLE)
            {
               KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_LEAVE);
               EventManager.postEvent(ki);
            }
            
            repaint(); 
         }
      }

      ThinClient.getInstance().getFocusManager().setFocusOn(this);
   }

   /**
    * Set the list's vertical scrollarr.
    * 
    * @param    vbar
    *           The vertical scrollbar.
    */
   public void setScrollBar(ScrollBar<GuiOutputManager> vbar)
   {
      this.vbar = vbar;
   }

   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    *          <code>TRUE</code> to make the widget enabled, 
    *          <code>FALSE</code> to disable the widget.
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      // set flag for main part
      super.setEnabled(enabled);

      // propagate the flag to vertical scrollbar
      if (vbar != null)
      {
         vbar.setEnabled(enabled);
      }
   }
   
   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return super.width() + 1.75;
   }

   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      
      // nothing to do in GUI mode 
   }
   
   /**
    * Intercept key events and handle some special cases which are not
    * handled by base class.
    * 
    * @param    kvt
    *           Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput kvt)
   {
      // disable ENTER key itself as event generator to avoid value change duplication
      // this event gas already registered as drop-down exit trigger
      if (kvt.keyCode() == Key.VK_ENTER)
      {
         kvt.consume();
      }
      else
      {
         super.processKeyEvent(kvt);
      }
   }

   /**
    * Check if the widget needs LEAVE event on special mouse press conditions.
    * 
    * @param    eventSrc
    *           The event source.
    * @param    evtToTest
    *           The mouse event to check.
    * 
    * @return   <code>TRUE</code> if a mouse action should cause the <code>LEAVE</code> event
    *           to be generated.
    */
   @Override
   public boolean isLeaveEventRequired(Widget<GuiOutputManager> eventSrc, MouseEvt evtToTest)
   {
      // the event source must be not null
      if (eventSrc != null)
      {
         // check for possible clicks inside drop-down but not in the list itself
         // it can be the drop-down buttons for example
         Widget<GuiOutputManager> wAncest = eventSrc.ancestor();

         // if we are going outside the drop-down - send leave to it
         if (evtToTest.getEvent().getID() == MouseEvt.MOUSE_PRESSED &&
             (wAncest == null || !(wAncest instanceof DropDownGuiImpl)))
         {
            return true;
         }
      }

      return false;
   }
   
   /**
    * Set item number to be displayed on the top of the drop-down list.
    * 
    * @param   index
    *          The index of the top item.
    */
   @Override
   protected void setTopRow(int index)
   {
      ensureRowVisible(index, false);
      rowTop = index;
   }
   
   /**
    * Draws the next text item in the list inside visible part of the drop-down.
    * 
    * @param    number
    *           The number of the item to draw.
    * @param    x
    *           The X coordinate of the enclosing rectangle to draw item.
    * @param    y
    *           The Y coordinate of the enclosing rectangle to draw item.
    * @param    rw
    *           The pixel width of the enclosing rectangle to draw item.
    * @param    rh
    *           The pixel height of the enclosing rectangle to draw item.
    * @param    fontSize
    *           The vertical size of font to draw item text.
    * @param    enabled
    *           Flag indicating the enabled state of the list.
    */
   private void drawItem(int number, int x, int y, int rw, int rh, int fontSize, boolean enabled)
   {
      if (number == currentRow())
      {
         gd.setColor(selectionBack);
         gd.fillRect(x, y, rw, rh);
      }

      if (enabled && number == currentRow())
      {
         gd.setColor(selectionFore);
      }
      else if (enabled)
      {
         gd.setColor(gc.fgColor);
      }
      else
      {
         gd.setColor(disabledFore);
      }

      gd.drawString(getItemText(number), x + ComboBoxGuiImpl.TEXT_SHIFT, y + (rh + fontSize)/2);
   }
}
